aboutsummaryrefslogtreecommitdiffstats
path: root/whisper/whisperv5
diff options
context:
space:
mode:
Diffstat (limited to 'whisper/whisperv5')
-rw-r--r--whisper/whisperv5/api.go795
-rw-r--r--whisper/whisperv5/api_test.go680
-rw-r--r--whisper/whisperv5/config.go29
-rw-r--r--whisper/whisperv5/doc.go5
-rw-r--r--whisper/whisperv5/envelope.go2
-rw-r--r--whisper/whisperv5/filter.go1
-rw-r--r--whisper/whisperv5/filter_test.go4
-rw-r--r--whisper/whisperv5/gen_criteria_json.go62
-rw-r--r--whisper/whisperv5/gen_message_json.go80
-rw-r--r--whisper/whisperv5/gen_newmessage_json.go86
-rw-r--r--whisper/whisperv5/message.go23
-rw-r--r--whisper/whisperv5/peer_test.go2
-rw-r--r--whisper/whisperv5/topic.go32
-rw-r--r--whisper/whisperv5/topic_test.go70
-rw-r--r--whisper/whisperv5/whisper.go124
-rw-r--r--whisper/whisperv5/whisper_test.go48
16 files changed, 878 insertions, 1165 deletions
diff --git a/whisper/whisperv5/api.go b/whisper/whisperv5/api.go
index 1a4e4d879..5b84b99eb 100644
--- a/whisper/whisperv5/api.go
+++ b/whisper/whisperv5/api.go
@@ -17,494 +17,575 @@
package whisperv5
import (
- "encoding/json"
+ "context"
+ "crypto/ecdsa"
"errors"
"fmt"
+ "sync"
+ "time"
"github.com/ethereum/go-ethereum/common"
"github.com/ethereum/go-ethereum/common/hexutil"
"github.com/ethereum/go-ethereum/crypto"
+ "github.com/ethereum/go-ethereum/log"
"github.com/ethereum/go-ethereum/p2p/discover"
+ "github.com/ethereum/go-ethereum/rpc"
)
-var whisperOfflineErr = errors.New("whisper is offline")
+const (
+ filterTimeout = 300 // filters are considered timeout out after filterTimeout seconds
+)
+
+var (
+ ErrSymAsym = errors.New("specify either a symetric or a asymmetric key")
+ ErrInvalidSymmetricKey = errors.New("invalid symmetric key")
+ ErrInvalidPublicKey = errors.New("invalid public key")
+ ErrInvalidSigningPubKey = errors.New("invalid signing public key")
+ ErrTooLowPoW = errors.New("message rejected, PoW too low")
+ ErrNoTopics = errors.New("missing topic(s)")
+)
-// PublicWhisperAPI provides the whisper RPC service.
+// PublicWhisperAPI provides the whisper RPC service that can be
+// use publicly without security implications.
type PublicWhisperAPI struct {
- whisper *Whisper
+ w *Whisper
+
+ mu sync.Mutex
+ lastUsed map[string]time.Time // keeps track when a filter was polled for the last time.
}
// NewPublicWhisperAPI create a new RPC whisper service.
func NewPublicWhisperAPI(w *Whisper) *PublicWhisperAPI {
- return &PublicWhisperAPI{whisper: w}
+ api := &PublicWhisperAPI{
+ w: w,
+ lastUsed: make(map[string]time.Time),
+ }
+
+ go api.run()
+ return api
}
-// Start starts the Whisper worker threads.
-func (api *PublicWhisperAPI) Start() error {
- if api.whisper == nil {
- return whisperOfflineErr
+// run the api event loop.
+// this loop deletes filter that have not been used within filterTimeout
+func (api *PublicWhisperAPI) run() {
+ timeout := time.NewTicker(2 * time.Minute)
+ for {
+ <-timeout.C
+
+ api.mu.Lock()
+ for id, lastUsed := range api.lastUsed {
+ if time.Since(lastUsed).Seconds() >= filterTimeout {
+ delete(api.lastUsed, id)
+ if err := api.w.Unsubscribe(id); err != nil {
+ log.Error("could not unsubscribe whisper filter", "error", err)
+ }
+ log.Debug("delete whisper filter (timeout)", "id", id)
+ }
+ }
+ api.mu.Unlock()
}
- return api.whisper.Start(nil)
}
-// Stop stops the Whisper worker threads.
-func (api *PublicWhisperAPI) Stop() error {
- if api.whisper == nil {
- return whisperOfflineErr
- }
- return api.whisper.Stop()
+// Version returns the Whisper sub-protocol version.
+func (api *PublicWhisperAPI) Version(ctx context.Context) string {
+ return ProtocolVersionStr
}
-// Version returns the Whisper version this node offers.
-func (api *PublicWhisperAPI) Version() (hexutil.Uint, error) {
- if api.whisper == nil {
- return 0, whisperOfflineErr
- }
- return hexutil.Uint(api.whisper.Version()), nil
+// Info contains diagnostic information.
+type Info struct {
+ Memory int `json:"memory"` // Memory size of the floating messages in bytes.
+ Messages int `json:"messages"` // Number of floating messages.
+ MinPow float64 `json:"minPow"` // Minimal accepted PoW
+ MaxMessageSize uint32 `json:"maxMessageSize"` // Maximum accepted message size
}
-// Info returns the Whisper statistics for diagnostics.
-func (api *PublicWhisperAPI) Info() (string, error) {
- if api.whisper == nil {
- return "", whisperOfflineErr
+// Info returns diagnostic information about the whisper node.
+func (api *PublicWhisperAPI) Info(ctx context.Context) Info {
+ stats := api.w.Stats()
+ return Info{
+ Memory: stats.memoryUsed,
+ Messages: len(api.w.messageQueue) + len(api.w.p2pMsgQueue),
+ MinPow: api.w.MinPow(),
+ MaxMessageSize: api.w.MaxMessageSize(),
}
- return api.whisper.Stats(), nil
}
-// SetMaxMessageLength sets the maximal message length allowed by this node
-func (api *PublicWhisperAPI) SetMaxMessageLength(val int) error {
- if api.whisper == nil {
- return whisperOfflineErr
- }
- return api.whisper.SetMaxMessageLength(val)
+// SetMaxMessageSize sets the maximum message size that is accepted.
+// Upper limit is defined in whisperv5.MaxMessageSize.
+func (api *PublicWhisperAPI) SetMaxMessageSize(ctx context.Context, size uint32) (bool, error) {
+ return true, api.w.SetMaxMessageSize(size)
}
-// SetMinimumPoW sets the minimal PoW required by this node
-func (api *PublicWhisperAPI) SetMinimumPoW(val float64) error {
- if api.whisper == nil {
- return whisperOfflineErr
- }
- return api.whisper.SetMinimumPoW(val)
+// SetMinPow sets the minimum PoW for a message before it is accepted.
+func (api *PublicWhisperAPI) SetMinPoW(ctx context.Context, pow float64) (bool, error) {
+ return true, api.w.SetMinimumPoW(pow)
}
-// AllowP2PMessagesFromPeer marks specific peer trusted, which will allow it
-// to send historic (expired) messages.
-func (api *PublicWhisperAPI) AllowP2PMessagesFromPeer(enode string) error {
- if api.whisper == nil {
- return whisperOfflineErr
- }
+// MarkTrustedPeer marks a peer trusted. , which will allow it to send historic (expired) messages.
+// Note: This function is not adding new nodes, the node needs to exists as a peer.
+func (api *PublicWhisperAPI) MarkTrustedPeer(ctx context.Context, enode string) (bool, error) {
n, err := discover.ParseNode(enode)
if err != nil {
- return errors.New("failed to parse enode of trusted peer: " + err.Error())
+ return false, err
}
- return api.whisper.AllowP2PMessagesFromPeer(n.ID[:])
+ return true, api.w.AllowP2PMessagesFromPeer(n.ID[:])
}
-// HasKeyPair checks if the whisper node is configured with the private key
-// of the specified public pair.
-func (api *PublicWhisperAPI) HasKeyPair(id string) (bool, error) {
- if api.whisper == nil {
- return false, whisperOfflineErr
- }
- return api.whisper.HasKeyPair(id), nil
+// NewKeyPair generates a new public and private key pair for message decryption and encryption.
+// It returns an ID that can be used to refer to the keypair.
+func (api *PublicWhisperAPI) NewKeyPair(ctx context.Context) (string, error) {
+ return api.w.NewKeyPair()
}
-// DeleteKeyPair deletes the specifies key if it exists.
-func (api *PublicWhisperAPI) DeleteKeyPair(id string) (bool, error) {
- if api.whisper == nil {
- return false, whisperOfflineErr
+// AddPrivateKey imports the given private key.
+func (api *PublicWhisperAPI) AddPrivateKey(ctx context.Context, privateKey hexutil.Bytes) (string, error) {
+ key, err := crypto.ToECDSA(privateKey)
+ if err != nil {
+ return "", err
}
- return api.whisper.DeleteKeyPair(id), nil
+ return api.w.AddKeyPair(key)
}
-// NewKeyPair generates a new cryptographic identity for the client, and injects
-// it into the known identities for message decryption.
-func (api *PublicWhisperAPI) NewKeyPair() (string, error) {
- if api.whisper == nil {
- return "", whisperOfflineErr
+// DeleteKeyPair removes the key with the given key if it exists.
+func (api *PublicWhisperAPI) DeleteKeyPair(ctx context.Context, key string) (bool, error) {
+ if ok := api.w.DeleteKeyPair(key); ok {
+ return true, nil
}
- return api.whisper.NewKeyPair()
+ return false, fmt.Errorf("key pair %s not found", key)
}
-// GetPublicKey returns the public key for identity id
-func (api *PublicWhisperAPI) GetPublicKey(id string) (hexutil.Bytes, error) {
- if api.whisper == nil {
- return nil, whisperOfflineErr
- }
- key, err := api.whisper.GetPrivateKey(id)
+// HasKeyPair returns an indication if the node has a key pair that is associated with the given id.
+func (api *PublicWhisperAPI) HasKeyPair(ctx context.Context, id string) bool {
+ return api.w.HasKeyPair(id)
+}
+
+// GetPublicKey returns the public key associated with the given key. The key is the hex
+// encoded representation of a key in the form specified in section 4.3.6 of ANSI X9.62.
+func (api *PublicWhisperAPI) GetPublicKey(ctx context.Context, id string) (hexutil.Bytes, error) {
+ key, err := api.w.GetPrivateKey(id)
if err != nil {
- return nil, err
+ return hexutil.Bytes{}, err
}
return crypto.FromECDSAPub(&key.PublicKey), nil
}
-// GetPrivateKey returns the private key for identity id
-func (api *PublicWhisperAPI) GetPrivateKey(id string) (string, error) {
- if api.whisper == nil {
- return "", whisperOfflineErr
- }
- key, err := api.whisper.GetPrivateKey(id)
+// GetPublicKey returns the private key associated with the given key. The key is the hex
+// encoded representation of a key in the form specified in section 4.3.6 of ANSI X9.62.
+func (api *PublicWhisperAPI) GetPrivateKey(ctx context.Context, id string) (hexutil.Bytes, error) {
+ key, err := api.w.GetPrivateKey(id)
if err != nil {
- return "", err
+ return hexutil.Bytes{}, err
}
- return common.ToHex(crypto.FromECDSA(key)), nil
+ return crypto.FromECDSA(key), nil
}
-// GenerateSymmetricKey generates a random symmetric key and stores it under id,
-// which is then returned. Will be used in the future for session key exchange.
-func (api *PublicWhisperAPI) GenerateSymmetricKey() (string, error) {
- if api.whisper == nil {
- return "", whisperOfflineErr
- }
- return api.whisper.GenerateSymKey()
+// NewSymKey generate a random symmetric key.
+// It returns an ID that can be used to refer to the key.
+// Can be used encrypting and decrypting messages where the key is known to both parties.
+func (api *PublicWhisperAPI) NewSymKey(ctx context.Context) (string, error) {
+ return api.w.GenerateSymKey()
}
-// AddSymmetricKeyDirect stores the key, and returns its id.
-func (api *PublicWhisperAPI) AddSymmetricKeyDirect(key hexutil.Bytes) (string, error) {
- if api.whisper == nil {
- return "", whisperOfflineErr
- }
- return api.whisper.AddSymKeyDirect(key)
+// AddSymKey import a symmetric key.
+// It returns an ID that can be used to refer to the key.
+// Can be used encrypting and decrypting messages where the key is known to both parties.
+func (api *PublicWhisperAPI) AddSymKey(ctx context.Context, key hexutil.Bytes) (string, error) {
+ return api.w.AddSymKeyDirect([]byte(key))
}
-// AddSymmetricKeyFromPassword generates the key from password, stores it, and returns its id.
-func (api *PublicWhisperAPI) AddSymmetricKeyFromPassword(password string) (string, error) {
- if api.whisper == nil {
- return "", whisperOfflineErr
- }
- return api.whisper.AddSymKeyFromPassword(password)
+// GenerateSymKeyFromPassword derive a key from the given password, stores it, and returns its ID.
+func (api *PublicWhisperAPI) GenerateSymKeyFromPassword(ctx context.Context, passwd string) (string, error) {
+ return api.w.AddSymKeyFromPassword(passwd)
}
-// HasSymmetricKey returns true if there is a key associated with the given id.
-// Otherwise returns false.
-func (api *PublicWhisperAPI) HasSymmetricKey(id string) (bool, error) {
- if api.whisper == nil {
- return false, whisperOfflineErr
- }
- res := api.whisper.HasSymKey(id)
- return res, nil
+// HasSymKey returns an indication if the node has a symmetric key associated with the given key.
+func (api *PublicWhisperAPI) HasSymKey(ctx context.Context, id string) bool {
+ return api.w.HasSymKey(id)
}
-// GetSymmetricKey returns the symmetric key associated with the given id.
-func (api *PublicWhisperAPI) GetSymmetricKey(name string) (hexutil.Bytes, error) {
- if api.whisper == nil {
- return nil, whisperOfflineErr
- }
- b, err := api.whisper.GetSymKey(name)
- if err != nil {
- return nil, err
- }
- return b, nil
+// GetSymKey returns the symmetric key associated with the given id.
+func (api *PublicWhisperAPI) GetSymKey(ctx context.Context, id string) (hexutil.Bytes, error) {
+ return api.w.GetSymKey(id)
}
-// DeleteSymmetricKey deletes the key associated with the name string if it exists.
-func (api *PublicWhisperAPI) DeleteSymmetricKey(name string) (bool, error) {
- if api.whisper == nil {
- return false, whisperOfflineErr
- }
- res := api.whisper.DeleteSymKey(name)
- return res, nil
+// DeleteSymKey deletes the symmetric key that is associated with the given id.
+func (api *PublicWhisperAPI) DeleteSymKey(ctx context.Context, id string) bool {
+ return api.w.DeleteSymKey(id)
}
-// Subscribe creates and registers a new filter to watch for inbound whisper messages.
-// Returns the ID of the newly created filter.
-func (api *PublicWhisperAPI) Subscribe(args WhisperFilterArgs) (string, error) {
- if api.whisper == nil {
- return "", whisperOfflineErr
- }
+//go:generate gencodec -type NewMessage -field-override newMessageOverride -out gen_newmessage_json.go
+
+// NewMessage represents a new whisper message that is posted through the RPC.
+type NewMessage struct {
+ SymKeyID string `json:"symKeyID"`
+ PublicKey []byte `json:"pubKey"`
+ Sig string `json:"sig"`
+ TTL uint32 `json:"ttl"`
+ Topic TopicType `json:"topic"`
+ Payload []byte `json:"payload"`
+ Padding []byte `json:"padding"`
+ PowTime uint32 `json:"powTime"`
+ PowTarget float64 `json:"powTarget"`
+ TargetPeer string `json:"targetPeer"`
+}
- filter := Filter{
- PoW: args.MinPoW,
- Messages: make(map[common.Hash]*ReceivedMessage),
- AllowP2P: args.AllowP2P,
- }
+type newMessageOverride struct {
+ PublicKey hexutil.Bytes
+ Payload hexutil.Bytes
+ Padding hexutil.Bytes
+}
- var err error
- for i, bt := range args.Topics {
- if len(bt) == 0 || len(bt) > 4 {
- return "", errors.New(fmt.Sprintf("subscribe: topic %d has wrong size: %d", i, len(bt)))
- }
- filter.Topics = append(filter.Topics, bt)
+// Post a message on the Whisper network.
+func (api *PublicWhisperAPI) Post(ctx context.Context, req NewMessage) (bool, error) {
+ var (
+ symKeyGiven = len(req.SymKeyID) > 0
+ pubKeyGiven = len(req.PublicKey) > 0
+ err error
+ )
+
+ // user must specify either a symmetric or a asymmetric key
+ if (symKeyGiven && pubKeyGiven) || (!symKeyGiven && !pubKeyGiven) {
+ return false, ErrSymAsym
}
- if err = ValidateKeyID(args.Key); err != nil {
- return "", errors.New("subscribe: " + err.Error())
+ params := &MessageParams{
+ TTL: req.TTL,
+ Payload: req.Payload,
+ Padding: req.Padding,
+ WorkTime: req.PowTime,
+ PoW: req.PowTarget,
+ Topic: req.Topic,
}
- if len(args.Sig) > 0 {
- sb := common.FromHex(args.Sig)
- if sb == nil {
- return "", errors.New("subscribe: sig parameter is invalid")
- }
- filter.Src = crypto.ToECDSAPub(sb)
- if !ValidatePublicKey(filter.Src) {
- return "", errors.New("subscribe: invalid 'sig' field")
+ // Set key that is used to sign the message
+ if len(req.Sig) > 0 {
+ if params.Src, err = api.w.GetPrivateKey(req.Sig); err != nil {
+ return false, err
}
}
- if args.Symmetric {
- if len(args.Topics) == 0 {
- return "", errors.New("subscribe: at least one topic must be specified with symmetric encryption")
+ // Set symmetric key that is used to encrypt the message
+ if symKeyGiven {
+ if params.Topic == (TopicType{}) { // topics are mandatory with symmetric encryption
+ return false, ErrNoTopics
}
- symKey, err := api.whisper.GetSymKey(args.Key)
- if err != nil {
- return "", errors.New("subscribe: invalid key ID")
+ if params.KeySym, err = api.w.GetSymKey(req.SymKeyID); err != nil {
+ return false, err
}
- if !validateSymmetricKey(symKey) {
- return "", errors.New("subscribe: retrieved key is invalid")
- }
- filter.KeySym = symKey
- filter.SymKeyHash = crypto.Keccak256Hash(filter.KeySym)
- } else {
- filter.KeyAsym, err = api.whisper.GetPrivateKey(args.Key)
- if err != nil {
- return "", errors.New("subscribe: invalid key ID")
+ if !validateSymmetricKey(params.KeySym) {
+ return false, ErrInvalidSymmetricKey
}
- if filter.KeyAsym == nil {
- return "", errors.New("subscribe: non-existent identity provided")
+ }
+
+ // Set asymmetric key that is used to encrypt the message
+ if pubKeyGiven {
+ params.Dst = crypto.ToECDSAPub(req.PublicKey)
+ if !ValidatePublicKey(params.Dst) {
+ return false, ErrInvalidPublicKey
}
}
- return api.whisper.Subscribe(&filter)
-}
+ // encrypt and sent message
+ whisperMsg, err := NewSentMessage(params)
+ if err != nil {
+ return false, err
+ }
-// Unsubscribe disables and removes an existing filter.
-func (api *PublicWhisperAPI) Unsubscribe(id string) {
- api.whisper.Unsubscribe(id)
-}
+ env, err := whisperMsg.Wrap(params)
+ if err != nil {
+ return false, err
+ }
+
+ // send to specific node (skip PoW check)
+ if len(req.TargetPeer) > 0 {
+ n, err := discover.ParseNode(req.TargetPeer)
+ if err != nil {
+ return false, fmt.Errorf("failed to parse target peer: %s", err)
+ }
+ return true, api.w.SendP2PMessage(n.ID[:], env)
+ }
-// GetSubscriptionMessages retrieves all the new messages matched by the corresponding
-// subscription filter since the last retrieval.
-func (api *PublicWhisperAPI) GetNewSubscriptionMessages(id string) []*WhisperMessage {
- f := api.whisper.GetFilter(id)
- if f != nil {
- newMail := f.Retrieve()
- return toWhisperMessages(newMail)
+ // ensure that the message PoW meets the node's minimum accepted PoW
+ if req.PowTarget < api.w.MinPow() {
+ return false, ErrTooLowPoW
}
- return toWhisperMessages(nil)
+
+ return true, api.w.Send(env)
}
-// GetMessages retrieves all the floating messages that match a specific subscription filter.
-// It is likely to be called once per session, right after Subscribe call.
-func (api *PublicWhisperAPI) GetFloatingMessages(id string) []*WhisperMessage {
- all := api.whisper.Messages(id)
- return toWhisperMessages(all)
+//go:generate gencodec -type Criteria -field-override criteriaOverride -out gen_criteria_json.go
+
+// Criteria holds various filter options for inbound messages.
+type Criteria struct {
+ SymKeyID string `json:"symKeyID"`
+ PrivateKeyID string `json:"privateKeyID"`
+ Sig []byte `json:"sig"`
+ MinPow float64 `json:"minPow"`
+ Topics []TopicType `json:"topics"`
+ AllowP2P bool `json:"allowP2P"`
}
-// toWhisperMessages converts a Whisper message to a RPC whisper message.
-func toWhisperMessages(messages []*ReceivedMessage) []*WhisperMessage {
- msgs := make([]*WhisperMessage, len(messages))
- for i, msg := range messages {
- msgs[i] = NewWhisperMessage(msg)
- }
- return msgs
+type criteriaOverride struct {
+ Sig hexutil.Bytes
}
-// Post creates a whisper message and injects it into the network for distribution.
-func (api *PublicWhisperAPI) Post(args PostArgs) error {
- if api.whisper == nil {
- return whisperOfflineErr
+// Messages set up a subscription that fires events when messages arrive that match
+// the given set of criteria.
+func (api *PublicWhisperAPI) Messages(ctx context.Context, crit Criteria) (*rpc.Subscription, error) {
+ var (
+ symKeyGiven = len(crit.SymKeyID) > 0
+ pubKeyGiven = len(crit.PrivateKeyID) > 0
+ err error
+ )
+
+ // ensure that the RPC connection supports subscriptions
+ notifier, supported := rpc.NotifierFromContext(ctx)
+ if !supported {
+ return nil, rpc.ErrNotificationsUnsupported
}
- var err error
- params := MessageParams{
- TTL: args.TTL,
- WorkTime: args.PowTime,
- PoW: args.PowTarget,
- Payload: args.Payload,
- Padding: args.Padding,
+ // user must specify either a symmetric or a asymmetric key
+ if (symKeyGiven && pubKeyGiven) || (!symKeyGiven && !pubKeyGiven) {
+ return nil, ErrSymAsym
}
- if len(args.Key) == 0 {
- return errors.New("post: key is missing")
+ filter := Filter{
+ PoW: crit.MinPow,
+ Messages: make(map[common.Hash]*ReceivedMessage),
+ AllowP2P: crit.AllowP2P,
}
- if len(args.Sig) > 0 {
- params.Src, err = api.whisper.GetPrivateKey(args.Sig)
- if err != nil {
- return err
- }
- if params.Src == nil {
- return errors.New("post: empty identity")
+ if len(crit.Sig) > 0 {
+ filter.Src = crypto.ToECDSAPub(crit.Sig)
+ if !ValidatePublicKey(filter.Src) {
+ return nil, ErrInvalidSigningPubKey
}
}
- if len(args.Topic) == TopicLength {
- params.Topic = BytesToTopic(args.Topic)
- } else if len(args.Topic) != 0 {
- return errors.New(fmt.Sprintf("post: wrong topic size %d", len(args.Topic)))
+ for i, bt := range crit.Topics {
+ if len(bt) == 0 || len(bt) > 4 {
+ return nil, fmt.Errorf("subscribe: topic %d has wrong size: %d", i, len(bt))
+ }
+ filter.Topics = append(filter.Topics, bt[:])
}
- if args.Type == "sym" {
- if err = ValidateKeyID(args.Key); err != nil {
- return err
+ // listen for message that are encrypted with the given symmetric key
+ if symKeyGiven {
+ if len(filter.Topics) == 0 {
+ return nil, ErrNoTopics
}
- params.KeySym, err = api.whisper.GetSymKey(args.Key)
+ key, err := api.w.GetSymKey(crit.SymKeyID)
if err != nil {
- return err
+ return nil, err
}
- if !validateSymmetricKey(params.KeySym) {
- return errors.New("post: key for symmetric encryption is invalid")
- }
- if len(params.Topic) == 0 {
- return errors.New("post: topic is missing for symmetric encryption")
+ if !validateSymmetricKey(key) {
+ return nil, ErrInvalidSymmetricKey
}
- } else if args.Type == "asym" {
- kb := common.FromHex(args.Key)
- if kb == nil {
- return errors.New("post: public key for asymmetric encryption is invalid")
- }
- params.Dst = crypto.ToECDSAPub(kb)
- if !ValidatePublicKey(params.Dst) {
- return errors.New("post: public key for asymmetric encryption is invalid")
- }
- } else {
- return errors.New("post: wrong type (sym/asym)")
+ filter.KeySym = key
+ filter.SymKeyHash = crypto.Keccak256Hash(filter.KeySym)
}
- // encrypt and send
- message, err := NewSentMessage(&params)
- if err != nil {
- return err
+ // listen for messages that are encrypted with the given public key
+ if pubKeyGiven {
+ filter.KeyAsym, err = api.w.GetPrivateKey(crit.PrivateKeyID)
+ if err != nil || filter.KeyAsym == nil {
+ return nil, ErrInvalidPublicKey
+ }
}
- envelope, err := message.Wrap(&params)
+
+ id, err := api.w.Subscribe(&filter)
if err != nil {
- return err
- }
- if envelope.size() > api.whisper.maxMsgLength {
- return errors.New("post: message is too big")
+ return nil, err
}
- if len(args.TargetPeer) != 0 {
- n, err := discover.ParseNode(args.TargetPeer)
- if err != nil {
- return errors.New("post: failed to parse enode of target peer: " + err.Error())
- }
- return api.whisper.SendP2PMessage(n.ID[:], envelope)
- } else if args.PowTarget < api.whisper.minPoW {
- return errors.New("post: target PoW is less than minimum PoW, the message can not be sent")
- }
-
- return api.whisper.Send(envelope)
-}
-
-type PostArgs struct {
- Type string `json:"type"` // "sym"/"asym" (symmetric or asymmetric)
- TTL uint32 `json:"ttl"` // time-to-live in seconds
- Sig string `json:"sig"` // id of the signing key
- Key string `json:"key"` // key id (in case of sym) or public key (in case of asym)
- Topic hexutil.Bytes `json:"topic"` // topic (4 bytes)
- Padding hexutil.Bytes `json:"padding"` // optional padding bytes
- Payload hexutil.Bytes `json:"payload"` // payload to be encrypted
- PowTime uint32 `json:"powTime"` // maximal time in seconds to be spent on PoW
- PowTarget float64 `json:"powTarget"` // minimal PoW required for this message
- TargetPeer string `json:"targetPeer"` // peer id (for p2p message only)
-}
-
-type WhisperFilterArgs struct {
- Symmetric bool // encryption type
- Key string // id of the key to be used for decryption
- Sig string // public key of the sender to be verified
- MinPoW float64 // minimal PoW requirement
- Topics [][]byte // list of topics (up to 4 bytes each) to match
- AllowP2P bool // indicates wheather direct p2p messages are allowed for this filter
-}
-
-// UnmarshalJSON implements the json.Unmarshaler interface, invoked to convert a
-// JSON message blob into a WhisperFilterArgs structure.
-func (args *WhisperFilterArgs) UnmarshalJSON(b []byte) (err error) {
- // Unmarshal the JSON message and sanity check
- var obj struct {
- Type string `json:"type"`
- Key string `json:"key"`
- Sig string `json:"sig"`
- MinPoW float64 `json:"minPoW"`
- Topics []interface{} `json:"topics"`
- AllowP2P bool `json:"allowP2P"`
- }
- if err := json.Unmarshal(b, &obj); err != nil {
- return err
- }
-
- switch obj.Type {
- case "sym":
- args.Symmetric = true
- case "asym":
- args.Symmetric = false
- default:
- return errors.New("wrong type (sym/asym)")
- }
-
- args.Key = obj.Key
- args.Sig = obj.Sig
- args.MinPoW = obj.MinPoW
- args.AllowP2P = obj.AllowP2P
-
- // Construct the topic array
- if obj.Topics != nil {
- topics := make([]string, len(obj.Topics))
- for i, field := range obj.Topics {
- switch value := field.(type) {
- case string:
- topics[i] = value
- case nil:
- return fmt.Errorf("topic[%d] is empty", i)
- default:
- return fmt.Errorf("topic[%d] is not a string", i)
- }
- }
- topicsDecoded := make([][]byte, len(topics))
- for j, s := range topics {
- x := common.FromHex(s)
- if x == nil || len(x) > TopicLength {
- return fmt.Errorf("topic[%d] is invalid", j)
+ // create subscription and start waiting for message events
+ rpcSub := notifier.CreateSubscription()
+ go func() {
+ // for now poll internally, refactor whisper internal for channel support
+ ticker := time.NewTicker(250 * time.Millisecond)
+ defer ticker.Stop()
+
+ for {
+ select {
+ case <-ticker.C:
+ if filter := api.w.GetFilter(id); filter != nil {
+ for _, rpcMessage := range toMessage(filter.Retrieve()) {
+ if err := notifier.Notify(rpcSub.ID, rpcMessage); err != nil {
+ log.Error("Failed to send notification", "err", err)
+ }
+ }
+ }
+ case <-rpcSub.Err():
+ api.w.Unsubscribe(id)
+ return
+ case <-notifier.Closed():
+ api.w.Unsubscribe(id)
+ return
}
- topicsDecoded[j] = x
}
- args.Topics = topicsDecoded
- }
+ }()
- return nil
+ return rpcSub, nil
}
-// WhisperMessage is the RPC representation of a whisper message.
-type WhisperMessage struct {
- Topic string `json:"topic"`
- Payload string `json:"payload"`
- Padding string `json:"padding"`
- Src string `json:"sig"`
- Dst string `json:"recipientPublicKey"`
- Timestamp uint32 `json:"timestamp"`
- TTL uint32 `json:"ttl"`
- PoW float64 `json:"pow"`
- Hash string `json:"hash"`
+//go:generate gencodec -type Message -field-override messageOverride -out gen_message_json.go
+
+// Message is the RPC representation of a whisper message.
+type Message struct {
+ Sig []byte `json:"sig,omitempty"`
+ TTL uint32 `json:"ttl"`
+ Timestamp uint32 `json:"timestamp"`
+ Topic TopicType `json:"topic"`
+ Payload []byte `json:"payload"`
+ Padding []byte `json:"padding"`
+ PoW float64 `json:"pow"`
+ Hash []byte `json:"hash"`
+ Dst []byte `json:"recipientPublicKey,omitempty"`
}
-// NewWhisperMessage converts an internal message into an API version.
-func NewWhisperMessage(message *ReceivedMessage) *WhisperMessage {
- msg := WhisperMessage{
- Payload: common.ToHex(message.Payload),
- Padding: common.ToHex(message.Padding),
+type messageOverride struct {
+ Sig hexutil.Bytes
+ Payload hexutil.Bytes
+ Padding hexutil.Bytes
+ Hash hexutil.Bytes
+ Dst hexutil.Bytes
+}
+
+// ToWhisperMessage converts an internal message into an API version.
+func ToWhisperMessage(message *ReceivedMessage) *Message {
+ msg := Message{
+ Payload: message.Payload,
+ Padding: message.Padding,
Timestamp: message.Sent,
TTL: message.TTL,
PoW: message.PoW,
- Hash: common.ToHex(message.EnvelopeHash.Bytes()),
+ Hash: message.EnvelopeHash.Bytes(),
+ Topic: message.Topic,
}
- if len(message.Topic) == TopicLength {
- msg.Topic = common.ToHex(message.Topic[:])
- }
if message.Dst != nil {
b := crypto.FromECDSAPub(message.Dst)
if b != nil {
- msg.Dst = common.ToHex(b)
+ msg.Dst = b
}
}
+
if isMessageSigned(message.Raw[0]) {
b := crypto.FromECDSAPub(message.SigToPubKey())
if b != nil {
- msg.Src = common.ToHex(b)
+ msg.Sig = b
}
}
+
return &msg
}
+
+// toMessage converts a set of messages to its RPC representation.
+func toMessage(messages []*ReceivedMessage) []*Message {
+ msgs := make([]*Message, len(messages))
+ for i, msg := range messages {
+ msgs[i] = ToWhisperMessage(msg)
+ }
+ return msgs
+}
+
+// GetFilterMessages returns the messages that match the filter criteria and
+// are received between the last poll and now.
+func (api *PublicWhisperAPI) GetFilterMessages(id string) ([]*Message, error) {
+ api.mu.Lock()
+ f := api.w.GetFilter(id)
+ if f == nil {
+ api.mu.Unlock()
+ return nil, fmt.Errorf("filter not found")
+ }
+ api.lastUsed[id] = time.Now()
+ api.mu.Unlock()
+
+ receivedMessages := f.Retrieve()
+ messages := make([]*Message, 0, len(receivedMessages))
+ for _, msg := range receivedMessages {
+ messages = append(messages, ToWhisperMessage(msg))
+ }
+
+ return messages, nil
+}
+
+// DeleteMessageFilter deletes a filter.
+func (api *PublicWhisperAPI) DeleteMessageFilter(id string) (bool, error) {
+ api.mu.Lock()
+ defer api.mu.Unlock()
+
+ delete(api.lastUsed, id)
+ return true, api.w.Unsubscribe(id)
+}
+
+// NewMessageFilter creates a new filter that can be used to poll for
+// (new) messages that satisfy the given criteria.
+func (api *PublicWhisperAPI) NewMessageFilter(req Criteria) (string, error) {
+ var (
+ src *ecdsa.PublicKey
+ keySym []byte
+ keyAsym *ecdsa.PrivateKey
+ topics [][]byte
+
+ symKeyGiven = len(req.SymKeyID) > 0
+ asymKeyGiven = len(req.PrivateKeyID) > 0
+
+ err error
+ )
+
+ // user must specify either a symmetric or a asymmetric key
+ if (symKeyGiven && asymKeyGiven) || (!symKeyGiven && !asymKeyGiven) {
+ return "", ErrSymAsym
+ }
+
+ if len(req.Sig) > 0 {
+ src = crypto.ToECDSAPub(req.Sig)
+ if !ValidatePublicKey(src) {
+ return "", ErrInvalidSigningPubKey
+ }
+ }
+
+ if symKeyGiven {
+ if keySym, err = api.w.GetSymKey(req.SymKeyID); err != nil {
+ return "", err
+ }
+ if !validateSymmetricKey(keySym) {
+ return "", ErrInvalidSymmetricKey
+ }
+ }
+
+ if asymKeyGiven {
+ if keyAsym, err = api.w.GetPrivateKey(req.PrivateKeyID); err != nil {
+ return "", err
+ }
+ }
+
+ if len(req.Topics) > 0 {
+ topics = make([][]byte, 1)
+ for _, topic := range req.Topics {
+ topics = append(topics, topic[:])
+ }
+ }
+
+ f := &Filter{
+ Src: src,
+ KeySym: keySym,
+ KeyAsym: keyAsym,
+ PoW: req.MinPow,
+ AllowP2P: req.AllowP2P,
+ Topics: topics,
+ Messages: make(map[common.Hash]*ReceivedMessage),
+ }
+
+ id, err := api.w.Subscribe(f)
+ if err != nil {
+ return "", err
+ }
+
+ api.mu.Lock()
+ api.lastUsed[id] = time.Now()
+ api.mu.Unlock()
+
+ return id, nil
+}
diff --git a/whisper/whisperv5/api_test.go b/whisper/whisperv5/api_test.go
deleted file mode 100644
index a6d82d850..000000000
--- a/whisper/whisperv5/api_test.go
+++ /dev/null
@@ -1,680 +0,0 @@
-// Copyright 2016 The go-ethereum Authors
-// This file is part of the go-ethereum library.
-//
-// The go-ethereum library is free software: you can redistribute it and/or modify
-// it under the terms of the GNU Lesser General Public License as published by
-// the Free Software Foundation, either version 3 of the License, or
-// (at your option) any later version.
-//
-// The go-ethereum library is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-// GNU Lesser General Public License for more details.
-//
-// You should have received a copy of the GNU Lesser General Public License
-// along with the go-ethereum library. If not, see <http://www.gnu.org/licenses/>.
-
-package whisperv5
-
-import (
- "bytes"
- "encoding/json"
- "testing"
- "time"
-
- "github.com/ethereum/go-ethereum/common"
- "github.com/ethereum/go-ethereum/common/hexutil"
-)
-
-func TestBasic(t *testing.T) {
- var id string = "test"
- w := New()
- api := NewPublicWhisperAPI(w)
- if api == nil {
- t.Fatalf("failed to create API.")
- }
-
- ver, err := api.Version()
- if err != nil {
- t.Fatalf("failed generateFilter: %s.", err)
- }
-
- if uint64(ver) != ProtocolVersion {
- t.Fatalf("wrong version: %d.", ver)
- }
-
- mail := api.GetNewSubscriptionMessages("non-existent-id")
- if len(mail) != 0 {
- t.Fatalf("failed GetFilterChanges: premature result")
- }
-
- exist, err := api.HasKeyPair(id)
- if err != nil {
- t.Fatalf("failed initial HasIdentity: %s.", err)
- }
- if exist {
- t.Fatalf("failed initial HasIdentity: false positive.")
- }
-
- success, err := api.DeleteKeyPair(id)
- if err != nil {
- t.Fatalf("failed DeleteIdentity: %s.", err)
- }
- if success {
- t.Fatalf("deleted non-existing identity: false positive.")
- }
-
- pub, err := api.NewKeyPair()
- if err != nil {
- t.Fatalf("failed NewIdentity: %s.", err)
- }
- if len(pub) == 0 {
- t.Fatalf("failed NewIdentity: empty")
- }
-
- exist, err = api.HasKeyPair(pub)
- if err != nil {
- t.Fatalf("failed HasIdentity: %s.", err)
- }
- if !exist {
- t.Fatalf("failed HasIdentity: false negative.")
- }
-
- success, err = api.DeleteKeyPair(pub)
- if err != nil {
- t.Fatalf("failed to delete second identity: %s.", err)
- }
- if !success {
- t.Fatalf("failed to delete second identity.")
- }
-
- exist, err = api.HasKeyPair(pub)
- if err != nil {
- t.Fatalf("failed HasIdentity(): %s.", err)
- }
- if exist {
- t.Fatalf("failed HasIdentity(): false positive.")
- }
-
- id = "arbitrary text"
- id2 := "another arbitrary string"
-
- exist, err = api.HasSymmetricKey(id)
- if err != nil {
- t.Fatalf("failed HasSymKey: %s.", err)
- }
- if exist {
- t.Fatalf("failed HasSymKey: false positive.")
- }
-
- id, err = api.GenerateSymmetricKey()
- if err != nil {
- t.Fatalf("failed GenerateSymKey: %s.", err)
- }
-
- exist, err = api.HasSymmetricKey(id)
- if err != nil {
- t.Fatalf("failed HasSymKey(): %s.", err)
- }
- if !exist {
- t.Fatalf("failed HasSymKey(): false negative.")
- }
-
- const password = "some stuff here"
- id, err = api.AddSymmetricKeyFromPassword(password)
- if err != nil {
- t.Fatalf("failed AddSymKey: %s.", err)
- }
-
- id2, err = api.AddSymmetricKeyFromPassword(password)
- if err != nil {
- t.Fatalf("failed AddSymKey: %s.", err)
- }
-
- exist, err = api.HasSymmetricKey(id2)
- if err != nil {
- t.Fatalf("failed HasSymKey(id2): %s.", err)
- }
- if !exist {
- t.Fatalf("failed HasSymKey(id2): false negative.")
- }
-
- k1, err := api.GetSymmetricKey(id)
- if err != nil {
- t.Fatalf("failed GetSymKey(id): %s.", err)
- }
- k2, err := api.GetSymmetricKey(id2)
- if err != nil {
- t.Fatalf("failed GetSymKey(id2): %s.", err)
- }
-
- if !bytes.Equal(k1, k2) {
- t.Fatalf("installed keys are not equal")
- }
-
- exist, err = api.DeleteSymmetricKey(id)
- if err != nil {
- t.Fatalf("failed DeleteSymKey(id): %s.", err)
- }
- if !exist {
- t.Fatalf("failed DeleteSymKey(id): false negative.")
- }
-
- exist, err = api.HasSymmetricKey(id)
- if err != nil {
- t.Fatalf("failed HasSymKey(id): %s.", err)
- }
- if exist {
- t.Fatalf("failed HasSymKey(id): false positive.")
- }
-}
-
-func TestUnmarshalFilterArgs(t *testing.T) {
- s := []byte(`{
- "type":"sym",
- "key":"0x70c87d191324e6712a591f304b4eedef6ad9bb9d",
- "sig":"0x9b2055d370f73ec7d8a03e965129118dc8f5bf83",
- "minPoW":2.34,
- "topics":["0x00000000", "0x007f80ff", "0xff807f00", "0xf26e7779"],
- "allowP2P":true
- }`)
-
- var f WhisperFilterArgs
- err := f.UnmarshalJSON(s)
- if err != nil {
- t.Fatalf("failed UnmarshalJSON: %s.", err)
- }
-
- if !f.Symmetric {
- t.Fatalf("wrong type.")
- }
- if f.Key != "0x70c87d191324e6712a591f304b4eedef6ad9bb9d" {
- t.Fatalf("wrong key: %s.", f.Key)
- }
- if f.Sig != "0x9b2055d370f73ec7d8a03e965129118dc8f5bf83" {
- t.Fatalf("wrong sig: %s.", f.Sig)
- }
- if f.MinPoW != 2.34 {
- t.Fatalf("wrong MinPoW: %f.", f.MinPoW)
- }
- if !f.AllowP2P {
- t.Fatalf("wrong AllowP2P.")
- }
- if len(f.Topics) != 4 {
- t.Fatalf("wrong topics number: %d.", len(f.Topics))
- }
-
- i := 0
- if !bytes.Equal(f.Topics[i], []byte{0x00, 0x00, 0x00, 0x00}) {
- t.Fatalf("wrong topic[%d]: %x.", i, f.Topics[i])
- }
-
- i++
- if !bytes.Equal(f.Topics[i], []byte{0x00, 0x7f, 0x80, 0xff}) {
- t.Fatalf("wrong topic[%d]: %x.", i, f.Topics[i])
- }
-
- i++
- if !bytes.Equal(f.Topics[i], []byte{0xff, 0x80, 0x7f, 0x00}) {
- t.Fatalf("wrong topic[%d]: %x.", i, f.Topics[i])
- }
-
- i++
- if !bytes.Equal(f.Topics[i], []byte{0xf2, 0x6e, 0x77, 0x79}) {
- t.Fatalf("wrong topic[%d]: %x.", i, f.Topics[i])
- }
-}
-
-func TestUnmarshalPostArgs(t *testing.T) {
- s := []byte(`{
- "type":"sym",
- "ttl":12345,
- "sig":"0x70c87d191324e6712a591f304b4eedef6ad9bb9d",
- "key":"0x9b2055d370f73ec7d8a03e965129118dc8f5bf83",
- "topic":"0xf26e7779",
- "padding":"0x74686973206973206D79207465737420737472696E67",
- "payload":"0x7061796C6F61642073686F756C642062652070736575646F72616E646F6D",
- "powTime":777,
- "powTarget":3.1416,
- "targetPeer":"enode://915533f667b1369793ebb9bda022416b1295235a1420799cd87a969467372546d808ebf59c5c9ce23f103d59b61b97df8af91f0908552485975397181b993461@127.0.0.1:12345"
- }`)
-
- var a PostArgs
- err := json.Unmarshal(s, &a)
- if err != nil {
- t.Fatalf("failed UnmarshalJSON: %s.", err)
- }
-
- if a.Type != "sym" {
- t.Fatalf("wrong Type: %s.", a.Type)
- }
- if a.TTL != 12345 {
- t.Fatalf("wrong ttl: %d.", a.TTL)
- }
- if a.Sig != "0x70c87d191324e6712a591f304b4eedef6ad9bb9d" {
- t.Fatalf("wrong From: %s.", a.Sig)
- }
- if a.Key != "0x9b2055d370f73ec7d8a03e965129118dc8f5bf83" {
- t.Fatalf("wrong Key: %s.", a.Key)
- }
-
- if BytesToTopic(a.Topic) != (TopicType{0xf2, 0x6e, 0x77, 0x79}) {
- t.Fatalf("wrong topic: %x.", a.Topic)
- }
- if string(a.Padding) != "this is my test string" {
- t.Fatalf("wrong Padding: %s.", string(a.Padding))
- }
- if string(a.Payload) != "payload should be pseudorandom" {
- t.Fatalf("wrong Payload: %s.", string(a.Payload))
- }
- if a.PowTime != 777 {
- t.Fatalf("wrong PowTime: %d.", a.PowTime)
- }
- if a.PowTarget != 3.1416 {
- t.Fatalf("wrong PowTarget: %f.", a.PowTarget)
- }
- if a.TargetPeer != "enode://915533f667b1369793ebb9bda022416b1295235a1420799cd87a969467372546d808ebf59c5c9ce23f103d59b61b97df8af91f0908552485975397181b993461@127.0.0.1:12345" {
- t.Fatalf("wrong PeerID: %s.", a.TargetPeer)
- }
-}
-
-func waitForMessages(api *PublicWhisperAPI, id string, target int) []*WhisperMessage {
- // timeout: 2 seconds
- result := make([]*WhisperMessage, 0, target)
- for i := 0; i < 100; i++ {
- mail := api.GetNewSubscriptionMessages(id)
- if len(mail) > 0 {
- for _, m := range mail {
- result = append(result, m)
- }
- if len(result) >= target {
- break
- }
- }
- time.Sleep(time.Millisecond * 20)
- }
-
- return result
-}
-
-func TestIntegrationAsym(t *testing.T) {
- w := New()
- api := NewPublicWhisperAPI(w)
- if api == nil {
- t.Fatalf("failed to create API.")
- }
-
- api.Start()
- defer api.Stop()
-
- sig, err := api.NewKeyPair()
- if err != nil {
- t.Fatalf("failed NewIdentity: %s.", err)
- }
- if len(sig) == 0 {
- t.Fatalf("wrong signature")
- }
-
- exist, err := api.HasKeyPair(sig)
- if err != nil {
- t.Fatalf("failed HasIdentity: %s.", err)
- }
- if !exist {
- t.Fatalf("failed HasIdentity: false negative.")
- }
-
- sigPubKey, err := api.GetPublicKey(sig)
- if err != nil {
- t.Fatalf("failed GetPublicKey: %s.", err)
- }
-
- key, err := api.NewKeyPair()
- if err != nil {
- t.Fatalf("failed NewIdentity(): %s.", err)
- }
- if len(key) == 0 {
- t.Fatalf("wrong key")
- }
-
- dstPubKey, err := api.GetPublicKey(key)
- if err != nil {
- t.Fatalf("failed GetPublicKey: %s.", err)
- }
-
- var topics [2]TopicType
- topics[0] = TopicType{0x00, 0x64, 0x00, 0xff}
- topics[1] = TopicType{0xf2, 0x6e, 0x77, 0x79}
- var f WhisperFilterArgs
- f.Symmetric = false
- f.Key = key
- f.Sig = sigPubKey.String()
- f.Topics = make([][]byte, 2)
- f.Topics[0] = topics[0][:]
- f.Topics[1] = topics[1][:]
- f.MinPoW = DefaultMinimumPoW / 2
- f.AllowP2P = true
-
- id, err := api.Subscribe(f)
- if err != nil {
- t.Fatalf("failed to create new filter: %s.", err)
- }
-
- var p PostArgs
- p.Type = "asym"
- p.TTL = 2
- p.Sig = sig
- p.Key = dstPubKey.String()
- p.Padding = []byte("test string")
- p.Payload = []byte("extended test string")
- p.PowTarget = DefaultMinimumPoW
- p.PowTime = 2
- p.Topic = hexutil.Bytes{0xf2, 0x6e, 0x77, 0x79} // topics[1]
-
- err = api.Post(p)
- if err != nil {
- t.Errorf("failed to post message: %s.", err)
- }
-
- mail := waitForMessages(api, id, 1)
- if len(mail) != 1 {
- t.Fatalf("failed to GetFilterChanges: got %d messages.", len(mail))
- }
-
- text := string(common.FromHex(mail[0].Payload))
- if text != string("extended test string") {
- t.Fatalf("failed to decrypt first message: %s.", text)
- }
-
- p.Padding = []byte("new value")
- p.Payload = []byte("extended new value")
- err = api.Post(p)
- if err != nil {
- t.Fatalf("failed to post next message: %s.", err)
- }
-
- mail = waitForMessages(api, id, 1)
- if len(mail) != 1 {
- t.Fatalf("failed to GetFilterChanges: got %d messages.", len(mail))
- }
-
- text = string(common.FromHex(mail[0].Payload))
- if text != string("extended new value") {
- t.Fatalf("failed to decrypt second message: %s.", text)
- }
-}
-
-func TestIntegrationSym(t *testing.T) {
- w := New()
- api := NewPublicWhisperAPI(w)
- if api == nil {
- t.Fatalf("failed to create API.")
- }
-
- api.Start()
- defer api.Stop()
-
- symKeyID, err := api.GenerateSymmetricKey()
- if err != nil {
- t.Fatalf("failed GenerateSymKey: %s.", err)
- }
-
- sig, err := api.NewKeyPair()
- if err != nil {
- t.Fatalf("failed NewKeyPair: %s.", err)
- }
- if len(sig) == 0 {
- t.Fatalf("wrong signature")
- }
-
- sigPubKey, err := api.GetPublicKey(sig)
- if err != nil {
- t.Fatalf("failed GetPublicKey: %s.", err)
- }
-
- exist, err := api.HasKeyPair(sig)
- if err != nil {
- t.Fatalf("failed HasIdentity: %s.", err)
- }
- if !exist {
- t.Fatalf("failed HasIdentity: false negative.")
- }
-
- var topics [2]TopicType
- topics[0] = TopicType{0x00, 0x7f, 0x80, 0xff}
- topics[1] = TopicType{0xf2, 0x6e, 0x77, 0x79}
- var f WhisperFilterArgs
- f.Symmetric = true
- f.Key = symKeyID
- f.Topics = make([][]byte, 2)
- f.Topics[0] = topics[0][:]
- f.Topics[1] = topics[1][:]
- f.MinPoW = DefaultMinimumPoW / 2
- f.Sig = sigPubKey.String()
- f.AllowP2P = false
-
- id, err := api.Subscribe(f)
- if err != nil {
- t.Fatalf("failed to create new filter: %s.", err)
- }
-
- var p PostArgs
- p.Type = "sym"
- p.TTL = 1
- p.Key = symKeyID
- p.Sig = sig
- p.Padding = []byte("test string")
- p.Payload = []byte("extended test string")
- p.PowTarget = DefaultMinimumPoW
- p.PowTime = 2
- p.Topic = hexutil.Bytes{0xf2, 0x6e, 0x77, 0x79}
-
- err = api.Post(p)
- if err != nil {
- t.Fatalf("failed to post first message: %s.", err)
- }
-
- mail := waitForMessages(api, id, 1)
- if len(mail) != 1 {
- t.Fatalf("failed GetFilterChanges: got %d messages.", len(mail))
- }
-
- text := string(common.FromHex(mail[0].Payload))
- if text != string("extended test string") {
- t.Fatalf("failed to decrypt first message: %s.", text)
- }
-
- p.Padding = []byte("new value")
- p.Payload = []byte("extended new value")
- err = api.Post(p)
- if err != nil {
- t.Fatalf("failed to post second message: %s.", err)
- }
-
- mail = waitForMessages(api, id, 1)
- if len(mail) != 1 {
- t.Fatalf("failed second GetFilterChanges: got %d messages.", len(mail))
- }
-
- text = string(common.FromHex(mail[0].Payload))
- if text != string("extended new value") {
- t.Fatalf("failed to decrypt second message: %s.", text)
- }
-}
-
-func TestIntegrationSymWithFilter(t *testing.T) {
- w := New()
- api := NewPublicWhisperAPI(w)
- if api == nil {
- t.Fatalf("failed to create API.")
- }
-
- api.Start()
- defer api.Stop()
-
- symKeyID, err := api.GenerateSymmetricKey()
- if err != nil {
- t.Fatalf("failed to GenerateSymKey: %s.", err)
- }
-
- sigKeyID, err := api.NewKeyPair()
- if err != nil {
- t.Fatalf("failed NewIdentity: %s.", err)
- }
- if len(sigKeyID) == 0 {
- t.Fatalf("wrong signature.")
- }
-
- exist, err := api.HasKeyPair(sigKeyID)
- if err != nil {
- t.Fatalf("failed HasIdentity: %s.", err)
- }
- if !exist {
- t.Fatalf("failed HasIdentity: does not exist.")
- }
-
- sigPubKey, err := api.GetPublicKey(sigKeyID)
- if err != nil {
- t.Fatalf("failed GetPublicKey: %s.", err)
- }
-
- var topics [2]TopicType
- topics[0] = TopicType{0x00, 0x7f, 0x80, 0xff}
- topics[1] = TopicType{0xf2, 0x6e, 0x77, 0x79}
- var f WhisperFilterArgs
- f.Symmetric = true
- f.Key = symKeyID
- f.Topics = make([][]byte, 2)
- f.Topics[0] = topics[0][:]
- f.Topics[1] = topics[1][:]
- f.MinPoW = DefaultMinimumPoW / 2
- f.Sig = sigPubKey.String()
- f.AllowP2P = false
-
- id, err := api.Subscribe(f)
- if err != nil {
- t.Fatalf("failed to create new filter: %s.", err)
- }
-
- var p PostArgs
- p.Type = "sym"
- p.TTL = 1
- p.Key = symKeyID
- p.Sig = sigKeyID
- p.Padding = []byte("test string")
- p.Payload = []byte("extended test string")
- p.PowTarget = DefaultMinimumPoW
- p.PowTime = 2
- p.Topic = hexutil.Bytes{0xf2, 0x6e, 0x77, 0x79}
-
- err = api.Post(p)
- if err != nil {
- t.Fatalf("failed to post message: %s.", err)
- }
-
- mail := waitForMessages(api, id, 1)
- if len(mail) != 1 {
- t.Fatalf("failed to GetFilterChanges: got %d messages.", len(mail))
- }
-
- text := string(common.FromHex(mail[0].Payload))
- if text != string("extended test string") {
- t.Fatalf("failed to decrypt first message: %s.", text)
- }
-
- p.Padding = []byte("new value")
- p.Payload = []byte("extended new value")
- err = api.Post(p)
- if err != nil {
- t.Fatalf("failed to post next message: %s.", err)
- }
-
- mail = waitForMessages(api, id, 1)
- if len(mail) != 1 {
- t.Fatalf("failed to GetFilterChanges: got %d messages.", len(mail))
- }
-
- text = string(common.FromHex(mail[0].Payload))
- if text != string("extended new value") {
- t.Fatalf("failed to decrypt second message: %s.", text)
- }
-}
-
-func TestKey(t *testing.T) {
- w := New()
- api := NewPublicWhisperAPI(w)
- if api == nil {
- t.Fatalf("failed to create API.")
- }
-
- k, err := api.AddSymmetricKeyFromPassword("wwww")
- if err != nil {
- t.Fatalf("failed to create key: %s.", err)
- }
-
- s, err := api.GetSymmetricKey(k)
- if err != nil {
- t.Fatalf("failed to get sym key: %s.", err)
- }
-
- k2, err := api.AddSymmetricKeyDirect(s)
- if err != nil {
- t.Fatalf("failed to add sym key: %s.", err)
- }
-
- s2, err := api.GetSymmetricKey(k2)
- if err != nil {
- t.Fatalf("failed to get sym key: %s.", err)
- }
-
- if s.String() != "0x448652d595bd6ec00b2a9ea220ad6c26592d9bf4cf79023d3c1b30cb681e6e07" {
- t.Fatalf("wrong key from password: %s", s.String())
- }
-
- if !bytes.Equal(s, s2) {
- t.Fatalf("wrong key")
- }
-}
-
-func TestSubscribe(t *testing.T) {
- var err error
- var s string
-
- w := New()
- api := NewPublicWhisperAPI(w)
- if api == nil {
- t.Fatalf("failed to create API.")
- }
-
- symKeyID, err := api.GenerateSymmetricKey()
- if err != nil {
- t.Fatalf("failed to GenerateSymKey: %s.", err)
- }
-
- var f WhisperFilterArgs
- f.Symmetric = true
- f.Key = symKeyID
- f.Topics = make([][]byte, 5)
- f.Topics[0] = []byte{0x21}
- f.Topics[1] = []byte{0xd2, 0xe3}
- f.Topics[2] = []byte{0x64, 0x75, 0x76}
- f.Topics[3] = []byte{0xf8, 0xe9, 0xa0, 0xba}
- f.Topics[4] = []byte{0xcb, 0x3c, 0xdd, 0xee, 0xff}
-
- s, err = api.Subscribe(f)
- if err == nil {
- t.Fatalf("Subscribe: false positive.")
- }
-
- f.Topics[4] = []byte{}
- if err == nil {
- t.Fatalf("Subscribe: false positive again.")
- }
-
- f.Topics[4] = []byte{0x00}
- s, err = api.Subscribe(f)
- if err != nil {
- t.Fatalf("failed to subscribe: %s.", err)
- } else {
- api.Unsubscribe(s)
- }
-}
diff --git a/whisper/whisperv5/config.go b/whisper/whisperv5/config.go
new file mode 100644
index 000000000..290bf8962
--- /dev/null
+++ b/whisper/whisperv5/config.go
@@ -0,0 +1,29 @@
+// Copyright 2017 The go-ethereum Authors
+// This file is part of the go-ethereum library.
+//
+// The go-ethereum library is free software: you can redistribute it and/or modify
+// it under the terms of the GNU Lesser General Public License as published by
+// the Free Software Foundation, either version 3 of the License, or
+// (at your option) any later version.
+//
+// The go-ethereum library is distributed in the hope that it will be useful,
+// but WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+// GNU Lesser General Public License for more details.
+//
+// You should have received a copy of the GNU Lesser General Public License
+// along with the go-ethereum library. If not, see <http://www.gnu.org/licenses/>.
+
+package whisperv5
+
+type Config struct {
+ MaxMessageSize uint32 `toml:",omitempty"`
+ MinimumAcceptedPOW float64 `toml:",omitempty"`
+}
+
+var DefaultConfig = Config{
+ MaxMessageSize: DefaultMaxMessageSize,
+ MinimumAcceptedPOW: DefaultMinimumPoW,
+}
+
+var ()
diff --git a/whisper/whisperv5/doc.go b/whisper/whisperv5/doc.go
index 768291a16..7a57488bd 100644
--- a/whisper/whisperv5/doc.go
+++ b/whisper/whisperv5/doc.go
@@ -55,8 +55,9 @@ const (
AESNonceLength = 12
keyIdSize = 32
- DefaultMaxMessageLength = 1024 * 1024
- DefaultMinimumPoW = 0.2
+ MaxMessageSize = uint32(10 * 1024 * 1024) // maximum accepted size of a message.
+ DefaultMaxMessageSize = uint32(1024 * 1024)
+ DefaultMinimumPoW = 0.2
padSizeLimit = 256 // just an arbitrary number, could be changed without breaking the protocol (must not exceed 2^24)
messageQueueLimit = 1024
diff --git a/whisper/whisperv5/envelope.go b/whisper/whisperv5/envelope.go
index d95fcab75..169cbba9d 100644
--- a/whisper/whisperv5/envelope.go
+++ b/whisper/whisperv5/envelope.go
@@ -62,7 +62,7 @@ func (e *Envelope) rlpWithoutNonce() []byte {
// NewEnvelope wraps a Whisper message with expiration and destination data
// included into an envelope for network forwarding.
-func NewEnvelope(ttl uint32, topic TopicType, aesNonce []byte, msg *SentMessage) *Envelope {
+func NewEnvelope(ttl uint32, topic TopicType, aesNonce []byte, msg *sentMessage) *Envelope {
env := Envelope{
Version: make([]byte, 1),
Expiry: uint32(time.Now().Add(time.Second * time.Duration(ttl)).Unix()),
diff --git a/whisper/whisperv5/filter.go b/whisper/whisperv5/filter.go
index 03101d4a4..d571160d7 100644
--- a/whisper/whisperv5/filter.go
+++ b/whisper/whisperv5/filter.go
@@ -163,6 +163,7 @@ func (f *Filter) Retrieve() (all []*ReceivedMessage) {
for _, msg := range f.Messages {
all = append(all, msg)
}
+
f.Messages = make(map[common.Hash]*ReceivedMessage) // delete old messages
return all
}
diff --git a/whisper/whisperv5/filter_test.go b/whisper/whisperv5/filter_test.go
index dd4ab9e8d..4ce87eee2 100644
--- a/whisper/whisperv5/filter_test.go
+++ b/whisper/whisperv5/filter_test.go
@@ -97,7 +97,7 @@ func TestInstallFilters(t *testing.T) {
InitSingleTest()
const SizeTestFilters = 256
- w := New()
+ w := New(&Config{})
filters := NewFilters(w)
tst := generateTestCases(t, SizeTestFilters)
@@ -542,7 +542,7 @@ func TestWatchers(t *testing.T) {
var x, firstID string
var err error
- w := New()
+ w := New(&Config{})
filters := NewFilters(w)
tst := generateTestCases(t, NumFilters)
for i = 0; i < NumFilters; i++ {
diff --git a/whisper/whisperv5/gen_criteria_json.go b/whisper/whisperv5/gen_criteria_json.go
new file mode 100644
index 000000000..8d3e9ee24
--- /dev/null
+++ b/whisper/whisperv5/gen_criteria_json.go
@@ -0,0 +1,62 @@
+// Code generated by github.com/fjl/gencodec. DO NOT EDIT.
+
+package whisperv5
+
+import (
+ "encoding/json"
+
+ "github.com/ethereum/go-ethereum/common/hexutil"
+)
+
+func (c Criteria) MarshalJSON() ([]byte, error) {
+ type Criteria struct {
+ SymKeyID string `json:"symKeyID"`
+ PrivateKeyID string `json:"privateKeyID"`
+ Sig hexutil.Bytes `json:"sig"`
+ MinPow float64 `json:"minPow"`
+ Topics []TopicType `json:"topics"`
+ AllowP2P bool `json:"allowP2P"`
+ }
+ var enc Criteria
+ enc.SymKeyID = c.SymKeyID
+ enc.PrivateKeyID = c.PrivateKeyID
+ enc.Sig = c.Sig
+ enc.MinPow = c.MinPow
+ enc.Topics = c.Topics
+ enc.AllowP2P = c.AllowP2P
+ return json.Marshal(&enc)
+}
+
+func (c *Criteria) UnmarshalJSON(input []byte) error {
+ type Criteria struct {
+ SymKeyID *string `json:"symKeyID"`
+ PrivateKeyID *string `json:"privateKeyID"`
+ Sig hexutil.Bytes `json:"sig"`
+ MinPow *float64 `json:"minPow"`
+ Topics []TopicType `json:"topics"`
+ AllowP2P *bool `json:"allowP2P"`
+ }
+ var dec Criteria
+ if err := json.Unmarshal(input, &dec); err != nil {
+ return err
+ }
+ if dec.SymKeyID != nil {
+ c.SymKeyID = *dec.SymKeyID
+ }
+ if dec.PrivateKeyID != nil {
+ c.PrivateKeyID = *dec.PrivateKeyID
+ }
+ if dec.Sig != nil {
+ c.Sig = dec.Sig
+ }
+ if dec.MinPow != nil {
+ c.MinPow = *dec.MinPow
+ }
+ if dec.Topics != nil {
+ c.Topics = dec.Topics
+ }
+ if dec.AllowP2P != nil {
+ c.AllowP2P = *dec.AllowP2P
+ }
+ return nil
+}
diff --git a/whisper/whisperv5/gen_message_json.go b/whisper/whisperv5/gen_message_json.go
new file mode 100644
index 000000000..26168225c
--- /dev/null
+++ b/whisper/whisperv5/gen_message_json.go
@@ -0,0 +1,80 @@
+// Code generated by github.com/fjl/gencodec. DO NOT EDIT.
+
+package whisperv5
+
+import (
+ "encoding/json"
+
+ "github.com/ethereum/go-ethereum/common/hexutil"
+)
+
+func (m Message) MarshalJSON() ([]byte, error) {
+ type Message struct {
+ Sig hexutil.Bytes `json:"sig,omitempty"`
+ TTL uint32 `json:"ttl"`
+ Timestamp uint32 `json:"timestamp"`
+ Topic TopicType `json:"topic"`
+ Payload hexutil.Bytes `json:"payload"`
+ Padding hexutil.Bytes `json:"padding"`
+ PoW float64 `json:"pow"`
+ Hash hexutil.Bytes `json:"hash"`
+ Dst hexutil.Bytes `json:"recipientPublicKey,omitempty"`
+ }
+ var enc Message
+ enc.Sig = m.Sig
+ enc.TTL = m.TTL
+ enc.Timestamp = m.Timestamp
+ enc.Topic = m.Topic
+ enc.Payload = m.Payload
+ enc.Padding = m.Padding
+ enc.PoW = m.PoW
+ enc.Hash = m.Hash
+ enc.Dst = m.Dst
+ return json.Marshal(&enc)
+}
+
+func (m *Message) UnmarshalJSON(input []byte) error {
+ type Message struct {
+ Sig hexutil.Bytes `json:"sig,omitempty"`
+ TTL *uint32 `json:"ttl"`
+ Timestamp *uint32 `json:"timestamp"`
+ Topic *TopicType `json:"topic"`
+ Payload hexutil.Bytes `json:"payload"`
+ Padding hexutil.Bytes `json:"padding"`
+ PoW *float64 `json:"pow"`
+ Hash hexutil.Bytes `json:"hash"`
+ Dst hexutil.Bytes `json:"recipientPublicKey,omitempty"`
+ }
+ var dec Message
+ if err := json.Unmarshal(input, &dec); err != nil {
+ return err
+ }
+ if dec.Sig != nil {
+ m.Sig = dec.Sig
+ }
+ if dec.TTL != nil {
+ m.TTL = *dec.TTL
+ }
+ if dec.Timestamp != nil {
+ m.Timestamp = *dec.Timestamp
+ }
+ if dec.Topic != nil {
+ m.Topic = *dec.Topic
+ }
+ if dec.Payload != nil {
+ m.Payload = dec.Payload
+ }
+ if dec.Padding != nil {
+ m.Padding = dec.Padding
+ }
+ if dec.PoW != nil {
+ m.PoW = *dec.PoW
+ }
+ if dec.Hash != nil {
+ m.Hash = dec.Hash
+ }
+ if dec.Dst != nil {
+ m.Dst = dec.Dst
+ }
+ return nil
+}
diff --git a/whisper/whisperv5/gen_newmessage_json.go b/whisper/whisperv5/gen_newmessage_json.go
new file mode 100644
index 000000000..0231cd919
--- /dev/null
+++ b/whisper/whisperv5/gen_newmessage_json.go
@@ -0,0 +1,86 @@
+// Code generated by github.com/fjl/gencodec. DO NOT EDIT.
+
+package whisperv5
+
+import (
+ "encoding/json"
+
+ "github.com/ethereum/go-ethereum/common/hexutil"
+)
+
+func (n NewMessage) MarshalJSON() ([]byte, error) {
+ type NewMessage struct {
+ SymKeyID string `json:"symKeyID"`
+ PublicKey hexutil.Bytes `json:"pubKey"`
+ Sig string `json:"sig"`
+ TTL uint32 `json:"ttl"`
+ Topic TopicType `json:"topic"`
+ Payload hexutil.Bytes `json:"payload"`
+ Padding hexutil.Bytes `json:"padding"`
+ PowTime uint32 `json:"powTime"`
+ PowTarget float64 `json:"powTarget"`
+ TargetPeer string `json:"targetPeer"`
+ }
+ var enc NewMessage
+ enc.SymKeyID = n.SymKeyID
+ enc.PublicKey = n.PublicKey
+ enc.Sig = n.Sig
+ enc.TTL = n.TTL
+ enc.Topic = n.Topic
+ enc.Payload = n.Payload
+ enc.Padding = n.Padding
+ enc.PowTime = n.PowTime
+ enc.PowTarget = n.PowTarget
+ enc.TargetPeer = n.TargetPeer
+ return json.Marshal(&enc)
+}
+
+func (n *NewMessage) UnmarshalJSON(input []byte) error {
+ type NewMessage struct {
+ SymKeyID *string `json:"symKeyID"`
+ PublicKey hexutil.Bytes `json:"pubKey"`
+ Sig *string `json:"sig"`
+ TTL *uint32 `json:"ttl"`
+ Topic *TopicType `json:"topic"`
+ Payload hexutil.Bytes `json:"payload"`
+ Padding hexutil.Bytes `json:"padding"`
+ PowTime *uint32 `json:"powTime"`
+ PowTarget *float64 `json:"powTarget"`
+ TargetPeer *string `json:"targetPeer"`
+ }
+ var dec NewMessage
+ if err := json.Unmarshal(input, &dec); err != nil {
+ return err
+ }
+ if dec.SymKeyID != nil {
+ n.SymKeyID = *dec.SymKeyID
+ }
+ if dec.PublicKey != nil {
+ n.PublicKey = dec.PublicKey
+ }
+ if dec.Sig != nil {
+ n.Sig = *dec.Sig
+ }
+ if dec.TTL != nil {
+ n.TTL = *dec.TTL
+ }
+ if dec.Topic != nil {
+ n.Topic = *dec.Topic
+ }
+ if dec.Payload != nil {
+ n.Payload = dec.Payload
+ }
+ if dec.Padding != nil {
+ n.Padding = dec.Padding
+ }
+ if dec.PowTime != nil {
+ n.PowTime = *dec.PowTime
+ }
+ if dec.PowTarget != nil {
+ n.PowTarget = *dec.PowTarget
+ }
+ if dec.TargetPeer != nil {
+ n.TargetPeer = *dec.TargetPeer
+ }
+ return nil
+}
diff --git a/whisper/whisperv5/message.go b/whisper/whisperv5/message.go
index 4ef469b51..c27535cd1 100644
--- a/whisper/whisperv5/message.go
+++ b/whisper/whisperv5/message.go
@@ -49,7 +49,7 @@ type MessageParams struct {
// SentMessage represents an end-user data packet to transmit through the
// Whisper protocol. These are wrapped into Envelopes that need not be
// understood by intermediate nodes, just forwarded.
-type SentMessage struct {
+type sentMessage struct {
Raw []byte
}
@@ -87,8 +87,8 @@ func (msg *ReceivedMessage) isAsymmetricEncryption() bool {
}
// NewMessage creates and initializes a non-signed, non-encrypted Whisper message.
-func NewSentMessage(params *MessageParams) (*SentMessage, error) {
- msg := SentMessage{}
+func NewSentMessage(params *MessageParams) (*sentMessage, error) {
+ msg := sentMessage{}
msg.Raw = make([]byte, 1, len(params.Payload)+len(params.Padding)+signatureLength+padSizeLimit)
msg.Raw[0] = 0 // set all the flags to zero
err := msg.appendPadding(params)
@@ -119,7 +119,7 @@ func intSize(i int) (s int) {
// appendPadding appends the pseudorandom padding bytes and sets the padding flag.
// The last byte contains the size of padding (thus, its size must not exceed 256).
-func (msg *SentMessage) appendPadding(params *MessageParams) error {
+func (msg *sentMessage) appendPadding(params *MessageParams) error {
rawSize := len(params.Payload) + 1
if params.Src != nil {
rawSize += signatureLength
@@ -164,7 +164,7 @@ func (msg *SentMessage) appendPadding(params *MessageParams) error {
// sign calculates and sets the cryptographic signature for the message,
// also setting the sign flag.
-func (msg *SentMessage) sign(key *ecdsa.PrivateKey) error {
+func (msg *sentMessage) sign(key *ecdsa.PrivateKey) error {
if isMessageSigned(msg.Raw[0]) {
// this should not happen, but no reason to panic
log.Error("failed to sign the message: already signed")
@@ -183,7 +183,7 @@ func (msg *SentMessage) sign(key *ecdsa.PrivateKey) error {
}
// encryptAsymmetric encrypts a message with a public key.
-func (msg *SentMessage) encryptAsymmetric(key *ecdsa.PublicKey) error {
+func (msg *sentMessage) encryptAsymmetric(key *ecdsa.PublicKey) error {
if !ValidatePublicKey(key) {
return errors.New("invalid public key provided for asymmetric encryption")
}
@@ -196,7 +196,7 @@ func (msg *SentMessage) encryptAsymmetric(key *ecdsa.PublicKey) error {
// encryptSymmetric encrypts a message with a topic key, using AES-GCM-256.
// nonce size should be 12 bytes (see cipher.gcmStandardNonceSize).
-func (msg *SentMessage) encryptSymmetric(key []byte) (nonce []byte, err error) {
+func (msg *sentMessage) encryptSymmetric(key []byte) (nonce []byte, err error) {
if !validateSymmetricKey(key) {
return nil, errors.New("invalid key provided for symmetric encryption")
}
@@ -224,13 +224,12 @@ func (msg *SentMessage) encryptSymmetric(key []byte) (nonce []byte, err error) {
}
// Wrap bundles the message into an Envelope to transmit over the network.
-func (msg *SentMessage) Wrap(options *MessageParams) (envelope *Envelope, err error) {
+func (msg *sentMessage) Wrap(options *MessageParams) (envelope *Envelope, err error) {
if options.TTL == 0 {
options.TTL = DefaultTTL
}
if options.Src != nil {
- err = msg.sign(options.Src)
- if err != nil {
+ if err = msg.sign(options.Src); err != nil {
return nil, err
}
}
@@ -242,14 +241,12 @@ func (msg *SentMessage) Wrap(options *MessageParams) (envelope *Envelope, err er
} else {
err = errors.New("unable to encrypt the message: neither symmetric nor assymmetric key provided")
}
-
if err != nil {
return nil, err
}
envelope = NewEnvelope(options.TTL, options.Topic, nonce, msg)
- err = envelope.Seal(options)
- if err != nil {
+ if err = envelope.Seal(options); err != nil {
return nil, err
}
return envelope, nil
diff --git a/whisper/whisperv5/peer_test.go b/whisper/whisperv5/peer_test.go
index d3cd63b0b..bae2adb6f 100644
--- a/whisper/whisperv5/peer_test.go
+++ b/whisper/whisperv5/peer_test.go
@@ -113,7 +113,7 @@ func initialize(t *testing.T) {
for i := 0; i < NumNodes; i++ {
var node TestNode
- node.shh = New()
+ node.shh = New(&DefaultConfig)
node.shh.SetMinimumPoW(0.00000001)
node.shh.Start(nil)
topics := make([]TopicType, 0)
diff --git a/whisper/whisperv5/topic.go b/whisper/whisperv5/topic.go
index 54d7422d1..d1996c460 100644
--- a/whisper/whisperv5/topic.go
+++ b/whisper/whisperv5/topic.go
@@ -19,10 +19,8 @@
package whisperv5
import (
- "fmt"
- "strings"
-
"github.com/ethereum/go-ethereum/common"
+ "github.com/ethereum/go-ethereum/common/hexutil"
)
// Topic represents a cryptographically secure, probabilistic partial
@@ -46,24 +44,12 @@ func (topic *TopicType) String() string {
return string(common.ToHex(topic[:]))
}
-// UnmarshalJSON parses a hex representation to a topic.
-func (t *TopicType) UnmarshalJSON(input []byte) error {
- length := len(input)
- if length >= 2 && input[0] == '"' && input[length-1] == '"' {
- input = input[1 : length-1]
- }
- // strip "0x" for length check
- if len(input) > 1 && strings.ToLower(string(input[:2])) == "0x" {
- input = input[2:]
- }
- // validate the length of the input
- if len(input) != TopicLength*2 {
- return fmt.Errorf("unmarshalJSON failed: topic must be exactly %d bytes", TopicLength)
- }
- b := common.FromHex(string(input))
- if b == nil {
- return fmt.Errorf("unmarshalJSON failed: wrong topic format")
- }
- *t = BytesToTopic(b)
- return nil
+// MarshalText returns the hex representation of t.
+func (t TopicType) MarshalText() ([]byte, error) {
+ return hexutil.Bytes(t[:]).MarshalText()
+}
+
+// UnmarshalText parses a hex representation to a topic.
+func (t *TopicType) UnmarshalText(input []byte) error {
+ return hexutil.UnmarshalFixedText("Topic", input, t[:])
}
diff --git a/whisper/whisperv5/topic_test.go b/whisper/whisperv5/topic_test.go
index df566da36..54bbeaf85 100644
--- a/whisper/whisperv5/topic_test.go
+++ b/whisper/whisperv5/topic_test.go
@@ -16,7 +16,10 @@
package whisperv5
-import "testing"
+import (
+ "encoding/json"
+ "testing"
+)
var topicStringTests = []struct {
topic TopicType
@@ -53,60 +56,55 @@ var bytesToTopicTests = []struct {
{topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: nil},
}
-func TestBytesToTopic(t *testing.T) {
- for i, tst := range bytesToTopicTests {
- top := BytesToTopic(tst.data)
- if top != tst.topic {
- t.Fatalf("failed test %d: have %v, want %v.", i, t, tst.topic)
- }
- }
-}
-
var unmarshalTestsGood = []struct {
topic TopicType
data []byte
}{
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("0x00000000")},
- {topic: TopicType{0x00, 0x7f, 0x80, 0xff}, data: []byte("0x007f80ff")},
- {topic: TopicType{0xff, 0x80, 0x7f, 0x00}, data: []byte("0xff807f00")},
- {topic: TopicType{0xf2, 0x6e, 0x77, 0x79}, data: []byte("0xf26e7779")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("00000000")},
- {topic: TopicType{0x00, 0x80, 0x01, 0x00}, data: []byte("00800100")},
- {topic: TopicType{0x00, 0x7f, 0x80, 0xff}, data: []byte("007f80ff")},
- {topic: TopicType{0xff, 0x80, 0x7f, 0x00}, data: []byte("ff807f00")},
- {topic: TopicType{0xf2, 0x6e, 0x77, 0x79}, data: []byte("f26e7779")},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"0x00000000"`)},
+ {topic: TopicType{0x00, 0x7f, 0x80, 0xff}, data: []byte(`"0x007f80ff"`)},
+ {topic: TopicType{0xff, 0x80, 0x7f, 0x00}, data: []byte(`"0xff807f00"`)},
+ {topic: TopicType{0xf2, 0x6e, 0x77, 0x79}, data: []byte(`"0xf26e7779"`)},
}
var unmarshalTestsBad = []struct {
topic TopicType
data []byte
}{
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("0x000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("0x0000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("0x000000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("0x0000000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("0000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("000000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("0000000000")},
- {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte("abcdefg0")},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"0x000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"0x0000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"0x000000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"0x0000000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"0000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"000000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"0000000000"`)},
+ {topic: TopicType{0x00, 0x00, 0x00, 0x00}, data: []byte(`"abcdefg0"`)},
}
var unmarshalTestsUgly = []struct {
topic TopicType
data []byte
}{
- {topic: TopicType{0x01, 0x00, 0x00, 0x00}, data: []byte("00000001")},
+ {topic: TopicType{0x01, 0x00, 0x00, 0x00}, data: []byte(`"0x00000001"`)},
+}
+
+func TestBytesToTopic(t *testing.T) {
+ for i, tst := range bytesToTopicTests {
+ top := BytesToTopic(tst.data)
+ if top != tst.topic {
+ t.Fatalf("failed test %d: have %v, want %v.", i, t, tst.topic)
+ }
+ }
}
func TestUnmarshalTestsGood(t *testing.T) {
for i, tst := range unmarshalTestsGood {
var top TopicType
- err := top.UnmarshalJSON(tst.data)
+ err := json.Unmarshal(tst.data, &top)
if err != nil {
- t.Fatalf("failed test %d. input: %v.", i, tst.data)
+ t.Errorf("failed test %d. input: %v. err: %v", i, tst.data, err)
} else if top != tst.topic {
- t.Fatalf("failed test %d: have %v, want %v.", i, t, tst.topic)
+ t.Errorf("failed test %d: have %v, want %v.", i, t, tst.topic)
}
}
}
@@ -115,7 +113,7 @@ func TestUnmarshalTestsBad(t *testing.T) {
// in this test UnmarshalJSON() is supposed to fail
for i, tst := range unmarshalTestsBad {
var top TopicType
- err := top.UnmarshalJSON(tst.data)
+ err := json.Unmarshal(tst.data, &top)
if err == nil {
t.Fatalf("failed test %d. input: %v.", i, tst.data)
}
@@ -126,11 +124,11 @@ func TestUnmarshalTestsUgly(t *testing.T) {
// in this test UnmarshalJSON() is NOT supposed to fail, but result should be wrong
for i, tst := range unmarshalTestsUgly {
var top TopicType
- err := top.UnmarshalJSON(tst.data)
+ err := json.Unmarshal(tst.data, &top)
if err != nil {
- t.Fatalf("failed test %d. input: %v.", i, tst.data)
+ t.Errorf("failed test %d. input: %v.", i, tst.data)
} else if top == tst.topic {
- t.Fatalf("failed test %d: have %v, want %v.", i, top, tst.topic)
+ t.Errorf("failed test %d: have %v, want %v.", i, top, tst.topic)
}
}
}
diff --git a/whisper/whisperv5/whisper_test.go b/whisper/whisperv5/whisper_test.go
index 225728c42..145143833 100644
--- a/whisper/whisperv5/whisper_test.go
+++ b/whisper/whisperv5/whisper_test.go
@@ -18,13 +18,14 @@ package whisperv5
import (
"bytes"
+ "crypto/ecdsa"
mrand "math/rand"
"testing"
"time"
)
func TestWhisperBasic(t *testing.T) {
- w := New()
+ w := New(&DefaultConfig)
p := w.Protocols()
shh := p[0]
if shh.Name != ProtocolName {
@@ -117,8 +118,39 @@ func TestWhisperBasic(t *testing.T) {
}
}
+func TestWhisperAsymmetricKeyImport(t *testing.T) {
+ var (
+ w = New(&DefaultConfig)
+ privateKeys []*ecdsa.PrivateKey
+ )
+
+ for i := 0; i < 50; i++ {
+ id, err := w.NewKeyPair()
+ if err != nil {
+ t.Fatalf("could not generate key: %v", err)
+ }
+
+ pk, err := w.GetPrivateKey(id)
+ if err != nil {
+ t.Fatalf("could not export private key: %v", err)
+ }
+
+ privateKeys = append(privateKeys, pk)
+
+ if !w.DeleteKeyPair(id) {
+ t.Fatalf("could not delete private key")
+ }
+ }
+
+ for _, pk := range privateKeys {
+ if _, err := w.AddKeyPair(pk); err != nil {
+ t.Fatalf("could not import private key: %v", err)
+ }
+ }
+}
+
func TestWhisperIdentityManagement(t *testing.T) {
- w := New()
+ w := New(&DefaultConfig)
id1, err := w.NewKeyPair()
if err != nil {
t.Fatalf("failed to generate new key pair: %s.", err)
@@ -240,7 +272,7 @@ func TestWhisperSymKeyManagement(t *testing.T) {
var err error
var k1, k2 []byte
- w := New()
+ w := New(&DefaultConfig)
id1 := string("arbitrary-string-1")
id2 := string("arbitrary-string-2")
@@ -443,7 +475,7 @@ func TestWhisperSymKeyManagement(t *testing.T) {
func TestExpiry(t *testing.T) {
InitSingleTest()
- w := New()
+ w := New(&DefaultConfig)
w.SetMinimumPoW(0.0000001)
defer w.SetMinimumPoW(DefaultMinimumPoW)
w.Start(nil)
@@ -500,9 +532,9 @@ func TestExpiry(t *testing.T) {
func TestCustomization(t *testing.T) {
InitSingleTest()
- w := New()
+ w := New(&DefaultConfig)
defer w.SetMinimumPoW(DefaultMinimumPoW)
- defer w.SetMaxMessageLength(DefaultMaxMessageLength)
+ defer w.SetMaxMessageSize(DefaultMaxMessageSize)
w.Start(nil)
defer w.Stop()
@@ -547,13 +579,13 @@ func TestCustomization(t *testing.T) {
if err != nil {
t.Fatalf("failed Wrap with seed %d: %s.", seed, err)
}
- w.SetMaxMessageLength(env.size() - 1)
+ w.SetMaxMessageSize(uint32(env.size() - 1))
err = w.Send(env)
if err == nil {
t.Fatalf("successfully sent oversized envelope (seed %d): false positive.", seed)
}
- w.SetMaxMessageLength(DefaultMaxMessageLength)
+ w.SetMaxMessageSize(DefaultMaxMessageSize)
err = w.Send(env)
if err != nil {
t.Fatalf("failed to send second envelope with seed %d: %s.", seed, err)
diff --git a/whisper/whisperv5/whisper.go b/whisper/whisperv5/whisper.go
index f2aad08ef..d09246f69 100644
--- a/whisper/whisperv5/whisper.go
+++ b/whisper/whisperv5/whisper.go
@@ -33,6 +33,7 @@ import (
"github.com/ethereum/go-ethereum/rpc"
"github.com/syndtr/goleveldb/leveldb/errors"
"golang.org/x/crypto/pbkdf2"
+ "golang.org/x/sync/syncmap"
set "gopkg.in/fatih/set.v0"
)
@@ -44,6 +45,12 @@ type Statistics struct {
totalMessagesCleared int
}
+const (
+ minPowIdx = iota // Minimal PoW required by the whisper node
+ maxMsgSizeIdx = iota // Maximal message length allowed by the whisper node
+ overflowIdx = iota // Indicator of message queue overflow
+)
+
// Whisper represents a dark communication interface through the Ethereum
// network, using its very own P2P communication layer.
type Whisper struct {
@@ -54,28 +61,31 @@ type Whisper struct {
symKeys map[string][]byte // Symmetric key storage
keyMu sync.RWMutex // Mutex associated with key storages
+ poolMu sync.RWMutex // Mutex to sync the message and expiration pools
envelopes map[common.Hash]*Envelope // Pool of envelopes currently tracked by this node
expirations map[uint32]*set.SetNonTS // Message expiration pool
- poolMu sync.RWMutex // Mutex to sync the message and expiration pools
- peers map[*Peer]struct{} // Set of currently active peers
peerMu sync.RWMutex // Mutex to sync the active peer set
+ peers map[*Peer]struct{} // Set of currently active peers
messageQueue chan *Envelope // Message queue for normal whisper messages
p2pMsgQueue chan *Envelope // Message queue for peer-to-peer messages (not to be forwarded any further)
quit chan struct{} // Channel used for graceful exit
- minPoW float64 // Minimal PoW required by the whisper node
- maxMsgLength int // Maximal message length allowed by the whisper node
- overflow bool // Indicator of message queue overflow
+ settings syncmap.Map // holds configuration settings that can be dynamically changed
- stats Statistics // Statistics of whisper node
+ statsMu sync.Mutex // guard stats
+ stats Statistics // Statistics of whisper node
mailServer MailServer // MailServer interface
}
// New creates a Whisper client ready to communicate through the Ethereum P2P network.
-func New() *Whisper {
+func New(cfg *Config) *Whisper {
+ if cfg == nil {
+ cfg = &DefaultConfig
+ }
+
whisper := &Whisper{
privateKeys: make(map[string]*ecdsa.PrivateKey),
symKeys: make(map[string][]byte),
@@ -85,22 +95,49 @@ func New() *Whisper {
messageQueue: make(chan *Envelope, messageQueueLimit),
p2pMsgQueue: make(chan *Envelope, messageQueueLimit),
quit: make(chan struct{}),
- minPoW: DefaultMinimumPoW,
- maxMsgLength: DefaultMaxMessageLength,
}
+
whisper.filters = NewFilters(whisper)
+ whisper.settings.Store(minPowIdx, cfg.MinimumAcceptedPOW)
+ whisper.settings.Store(maxMsgSizeIdx, cfg.MaxMessageSize)
+ whisper.settings.Store(overflowIdx, false)
+
// p2p whisper sub protocol handler
whisper.protocol = p2p.Protocol{
Name: ProtocolName,
Version: uint(ProtocolVersion),
Length: NumberOfMessageCodes,
Run: whisper.HandlePeer,
+ NodeInfo: func() interface{} {
+ return map[string]interface{}{
+ "version": ProtocolVersionStr,
+ "maxMessageSize": whisper.MaxMessageSize(),
+ "minimumPoW": whisper.MinPow(),
+ }
+ },
}
return whisper
}
+func (w *Whisper) MinPow() float64 {
+ val, _ := w.settings.Load(minPowIdx)
+ return val.(float64)
+}
+
+// MaxMessageSize returns the maximum accepted message size.
+func (w *Whisper) MaxMessageSize() uint32 {
+ val, _ := w.settings.Load(maxMsgSizeIdx)
+ return val.(uint32)
+}
+
+// Overflow returns an indication if the message queue is full.
+func (w *Whisper) Overflow() bool {
+ val, _ := w.settings.Load(overflowIdx)
+ return val.(bool)
+}
+
// APIs returns the RPC descriptors the Whisper implementation offers
func (w *Whisper) APIs() []rpc.API {
return []rpc.API{
@@ -129,12 +166,12 @@ func (w *Whisper) Version() uint {
return w.protocol.Version
}
-// SetMaxMessageLength sets the maximal message length allowed by this node
-func (w *Whisper) SetMaxMessageLength(val int) error {
- if val <= 0 {
- return fmt.Errorf("invalid message length: %d", val)
+// SetMaxMessageSize sets the maximal message size allowed by this node
+func (w *Whisper) SetMaxMessageSize(size uint32) error {
+ if size > MaxMessageSize {
+ return fmt.Errorf("message size too large [%d>%d]", size, MaxMessageSize)
}
- w.maxMsgLength = val
+ w.settings.Store(maxMsgSizeIdx, uint32(size))
return nil
}
@@ -143,7 +180,7 @@ func (w *Whisper) SetMinimumPoW(val float64) error {
if val <= 0.0 {
return fmt.Errorf("invalid PoW: %f", val)
}
- w.minPoW = val
+ w.settings.Store(minPowIdx, val)
return nil
}
@@ -240,6 +277,20 @@ func (w *Whisper) DeleteKeyPair(key string) bool {
return false
}
+// AddKeyPair imports a asymmetric private key and returns it identifier.
+func (w *Whisper) AddKeyPair(key *ecdsa.PrivateKey) (string, error) {
+ id, err := GenerateRandomID()
+ if err != nil {
+ return "", fmt.Errorf("failed to generate ID: %s", err)
+ }
+
+ w.keyMu.Lock()
+ w.privateKeys[id] = key
+ w.keyMu.Unlock()
+
+ return id, nil
+}
+
// HasKeyPair checks if the the whisper node is configured with the private key
// of the specified public pair.
func (w *Whisper) HasKeyPair(id string) bool {
@@ -451,7 +502,7 @@ func (wh *Whisper) runMessageLoop(p *Peer, rw p2p.MsgReadWriter) error {
log.Warn("message loop", "peer", p.peer.ID(), "err", err)
return err
}
- if packet.Size > uint32(wh.maxMsgLength) {
+ if packet.Size > wh.MaxMessageSize() {
log.Warn("oversized message received", "peer", p.peer.ID())
return errors.New("oversized message received")
}
@@ -532,7 +583,7 @@ func (wh *Whisper) add(envelope *Envelope) (bool, error) {
}
}
- if envelope.size() > wh.maxMsgLength {
+ if uint32(envelope.size()) > wh.MaxMessageSize() {
return false, fmt.Errorf("huge messages are not allowed [%x]", envelope.Hash())
}
@@ -547,7 +598,7 @@ func (wh *Whisper) add(envelope *Envelope) (bool, error) {
return false, fmt.Errorf("wrong size of AESNonce: %d bytes [env: %x]", aesNonceSize, envelope.Hash())
}
- if envelope.PoW() < wh.minPoW {
+ if envelope.PoW() < wh.MinPow() {
log.Debug("envelope with low PoW dropped", "PoW", envelope.PoW(), "hash", envelope.Hash().Hex())
return false, nil // drop envelope without error
}
@@ -571,7 +622,9 @@ func (wh *Whisper) add(envelope *Envelope) (bool, error) {
log.Trace("whisper envelope already cached", "hash", envelope.Hash().Hex())
} else {
log.Trace("cached whisper envelope", "hash", envelope.Hash().Hex())
+ wh.statsMu.Lock()
wh.stats.memoryUsed += envelope.size()
+ wh.statsMu.Unlock()
wh.postEvent(envelope, false) // notify the local node about the new message
if wh.mailServer != nil {
wh.mailServer.Archive(envelope)
@@ -600,13 +653,13 @@ func (w *Whisper) checkOverflow() {
queueSize := len(w.messageQueue)
if queueSize == messageQueueLimit {
- if !w.overflow {
- w.overflow = true
+ if !w.Overflow() {
+ w.settings.Store(overflowIdx, true)
log.Warn("message queue overflow")
}
} else if queueSize <= messageQueueLimit/2 {
- if w.overflow {
- w.overflow = false
+ if w.Overflow() {
+ w.settings.Store(overflowIdx, false)
log.Warn("message queue overflow fixed (back to normal)")
}
}
@@ -653,6 +706,8 @@ func (w *Whisper) expire() {
w.poolMu.Lock()
defer w.poolMu.Unlock()
+ w.statsMu.Lock()
+ defer w.statsMu.Unlock()
w.stats.reset()
now := uint32(time.Now().Unix())
for expiry, hashSet := range w.expirations {
@@ -673,17 +728,11 @@ func (w *Whisper) expire() {
}
// Stats returns the whisper node statistics.
-func (w *Whisper) Stats() string {
- result := fmt.Sprintf("Memory usage: %d bytes. Average messages cleared per expiry cycle: %d. Total messages cleared: %d.",
- w.stats.memoryUsed, w.stats.totalMessagesCleared/w.stats.cycles, w.stats.totalMessagesCleared)
- if w.stats.messagesCleared > 0 {
- result += fmt.Sprintf(" Latest expiry cycle cleared %d messages (%d bytes).",
- w.stats.messagesCleared, w.stats.memoryCleared)
- }
- if w.overflow {
- result += " Message queue state: overflow."
- }
- return result
+func (w *Whisper) Stats() Statistics {
+ w.statsMu.Lock()
+ defer w.statsMu.Unlock()
+
+ return w.stats
}
// Envelopes retrieves all the messages currently pooled by the node.
@@ -734,15 +783,6 @@ func (s *Statistics) reset() {
s.messagesCleared = 0
}
-// ValidateKeyID checks the format of key id.
-func ValidateKeyID(id string) error {
- const target = keyIdSize * 2
- if len(id) != target {
- return fmt.Errorf("wrong size of key ID (expected %d bytes, got %d)", target, len(id))
- }
- return nil
-}
-
// ValidatePublicKey checks the format of the given public key.
func ValidatePublicKey(k *ecdsa.PublicKey) bool {
return k != nil && k.X != nil && k.Y != nil && k.X.Sign() != 0 && k.Y.Sign() != 0